/**
 * @file singleton.cpp
 * @author Fabian Kantereit
 */

// Solution 1:
// very slow
class Singleton {
 private:
  Singleton() = default;
  ~Singleton() = default;

  static std::mutex m;
  static Singleton* instance_;

 public:
  Singleton* instance() {
    std::lock_guard<std::mutex> lock(m);

    if (instance_ == nullptr)
      instance_ = new Singleton();

    return instance_;
  }
};

// Solution 2:
// double checked locking pattern
// better but no longer recommended
// this might break due to compiler optimisation and instruction reordering
// see:
// http://www.aristeia.com/Papers/DDJ_Jul_Aug_2004_revised.pdf
class Singleton {
 private:
  Singleton() = default;
  ~Singleton() = default;

  static std::mutex m;
  static Singleton* instance_;
+ public:
  Singleton* instance() {
    if (instance_ == nullptr) { // 1st test
      std::lock_guard<std::mutex> lock(m);
      if (instance_ == nullptr) { // 2nd test
        instance_ = new Singleton();
      }
    }
    return instance_;
  }
};

// Best Solution if compiler is fully supports C++11:
class Singleton {
 private:
  Singleton() = default;
  ~Singleton() = default;

 public:
  Singleton& instance() {
    static Singleton instance_;
    return instance_;
  }
};
