/**
 * @file atomic_operations.cpp
 * @author Fabian Kantereit
 */

#include <mutex>
#include <thread>
#include <atomic>
#include <iostream>

const int limited = 10000000;

void unsafe_modify_variable(int& var, int difference) {
  for (int i = 0; i < limited; ++i) {
    var += difference;
  }
}

std::mutex m;

/**
 * simple stupid example for demonstration mutexes
 */
void use_mutex(int& var, int difference) {
  for (int i = 0; i < limited; ++i) {
    std::lock_guard<std::mutex> lock(m);
    var += difference;
  }
}

/**
 * use atomic operations for better performance
 */
void atomic_operation(std::atomic<int>& var, int difference) {
  for (int i = 0; i < limited; ++i) {
    var += difference;
  }
}

int main(int argc, char* argv[]) {
  int var = 0;

  auto time1 = std::chrono::high_resolution_clock::now();
  std::thread t1(unsafe_modify_variable, std::ref(var), 1);
  std::thread t2(unsafe_modify_variable, std::ref(var), -1);
  t1.join();
  t2.join();
  auto time_elapsed = chrono::high_resolution_clock::now() - time1;

  std::cout << "Value of var: " << var << " Time: "
            << std::chrono::duration_cast<std::chrono::milliseconds>(
                   time_elapsed).count() << "ms" << std::endl;

  int var2 = 0;

  time1 = std::chrono::high_resolution_clock::now();
  std::thread t3(use_mutex, std::ref(var2), 1);
  std::thread t4(use_mutex, std::ref(var2), -1);
  t3.join();
  t4.join();
  time_elapsed = std::chrono::high_resolution_clock::now() - time1;

  std::cout << "Value of var2: " << var2 << " Time: "
            << std::chrono::duration_cast<std::chrono::milliseconds>(
                   time_elapsed).count() << "ms" << std::endl;

  std::atomic<int> atomic_var(0);

  time1 = std::chrono::high_resolution_clock::now();
  std::thread t5(atomic_operation, std::ref(atomic_var), 1);
  std::thread t6(atomic_operation, std::ref(atomic_var), -1);
  t5.join();
  t6.join();
  time_elapsed = std::chrono::high_resolution_clock::now() - time1;

  std::cout << "Value of atomic_var: " << atomic_var << " Time: "
            << std::chrono::duration_cast<std::chrono::milliseconds>(
                   time_elapsed).count() << "ms" << std::endl;

  return 0;
}
